@@ -48,25 +48,22 @@ macro_rules! format_text {
     };
 }
 
-/// Provides a macro-based eDSL to structure your XML writing code.
-/// The `tag!` macro is defined within this macro.
+/// Writes an XML tag to a given `XmlWriter` and takes care of closing the tag.
 ///
 /// ## Basic usage example
 /// ```
 /// # use std::fmt::Error;
-/// # use xmlsafe::{XmlWriter, format_text, escape_text, with_writer};
+/// # use xmlsafe::{XmlWriter, format_text, escape_text, tag};
 /// #
 /// struct User {name: String, id: u8}
 ///
-/// fn list_users(w: XmlWriter, users: Vec<User>) -> Result<(), Error> {
-///     with_writer!(w, {
-///         tag!("ul", {
-///             for user in users {
-///                 tag!("li", "id"=user.id, {
-///                     w.write(escape_text(user.name))?;
-///                 });
-///             }
-///         });
+/// fn list_users(mut w: XmlWriter, users: Vec<User>) -> Result<(), Error> {
+///     tag!(w, "ul", {
+///         for user in users {
+///             tag!(w, "li", "id"=user.id, {
+///                 w.write(escape_text(user.name))?;
+///             });
+///         }
 ///     });
 ///     Ok(())
 /// }
@@ -84,17 +81,15 @@ macro_rules! format_text {
 ///
 /// ```
 /// # use std::fmt::Error;
-/// # use xmlsafe::{XmlWriter, with_writer};
+/// # use xmlsafe::{XmlWriter, tag};
 /// #
-/// fn generate(w: XmlWriter, max: u8) -> Result<(), Error> {
-///     with_writer!(w, {
-///         tag!("example", "data" << |value| {
-///             for i in 1..max + 1 {
-///                 value.write(i)?;
-///                 value.write(" ")?;
-///             }
-///             Ok(())
-///         });
+/// fn generate(mut w: XmlWriter, max: u8) -> Result<(), Error> {
+///     tag!(w, "example", "data" << |value| {
+///         for i in 1..max + 1 {
+///             value.write(i)?;
+///             value.write(" ")?;
+///         }
+///         Ok(())
 ///     });
 ///     Ok(())
 /// }
@@ -115,7 +110,7 @@ macro_rules! format_text {
 /// and content can be added via the `tag!` macro as usual. For example:
 ///
 /// ```
-/// # use xmlsafe::{AttWriter, NameSafe, TagConstructor, XmlWriter, with_writer};
+/// # use xmlsafe::{AttWriter, NameSafe, TagConstructor, XmlWriter, tag};
 /// # use std::fmt::Error;
 /// #
 /// struct Circle { center: Point, radius: f64 }
@@ -134,16 +129,15 @@ macro_rules! format_text {
 ///     }
 /// }
 ///
-/// fn generate(w: XmlWriter) -> Result<(), Error> {
-///     with_writer!(w, {
-///         tag!(
-///             Circle {
-///                 center: Point{ x: 3.0, y: 4.0 },
-///                 radius: 5.0,
-///             },
-///             "class" = "my-rectangle"
-///         );
-///     });
+/// fn generate(mut w: XmlWriter) -> Result<(), Error> {
+///     tag!(
+///         w,
+///         Circle {
+///             center: Point{ x: 3.0, y: 4.0 },
+///             radius: 5.0,
+///         },
+///         "class" = "my-rectangle"
+///     );
 ///     Ok(())
 /// }
 ///
@@ -152,63 +146,61 @@ macro_rules! format_text {
 /// assert_eq!(out, r#"<circle cx="3" cy="4" r="5" class="my-rectangle"/>"#);
 /// ```
 #[macro_export]
-macro_rules! with_writer {
-    // This macro is designed to play nice with rustfmt. For this reason :block
-    // is used for code that should be indented and the attributes are separated
-    // by commas.
-    ($writer:ident, $block:block) => {
+macro_rules! tag {
+    // This macro should play nice with rustfmt. Therefore :block is used for
+    // code that should be indented and the attributes are separated by commas.
+    // This macro should also work well with rust-analyzer, which is why
+    // the macro is split up into two rules instead of being more DRY.
+    ($writer:ident, $constructor:expr $(, $k:literal$t:tt$v:expr)*) => {
         {
-            // workaround for https://github.com/rust-lang/rust/issues/35853
-            macro_rules! with_dollar_sign {
-                ($a:tt => $b:tt) => {
-                    macro_rules! __with_dollar_sign { $a => $b }
-                    __with_dollar_sign!($);
-                }
+            let constructor: &dyn $crate::TagConstructor = &$constructor;
+            #[allow(unused_mut)]
+            let mut att_writer = constructor.attributes($writer.open_start_tag(constructor.tag_name())?)?;
+
+            #[allow(unused_macros)]
+            macro_rules! attr {
+                ($ak:literal << $av:expr) => {
+                    let mut att_val_writer = att_writer.key($ak)?;
+                    let fun: &dyn Fn(&mut $crate::AttValueWriter) -> Result<(), std::fmt::Error> = &$av;
+                    fun(&mut att_val_writer)?;
+                    att_writer = att_val_writer.end()?;
+                };
+                ($ak:literal=$av:expr) => {
+                    att_writer.write_attr($ak, $av)?;
+                };
             }
+            $(attr!($k$t$v);)*
 
-            let mut $writer = $writer;
-            with_dollar_sign! {
-                ($d:tt) => {
-                    macro_rules! tag {
-                        ($constructor:expr $d(, $k:literal$t:tt$v:expr)* $d(, $blk:block)?) => {
-                            let constructor: &dyn $crate::TagConstructor = &$constructor;
-                            #[allow(unused_mut)]
-                            let mut att_writer = constructor.attributes($writer.open_start_tag(constructor.tag_name())?)?;
-
-                            #[allow(unused_macros)]
-                            macro_rules! attr {
-                                ($ak:literal << $av:expr) => {
-                                    let mut att_val_writer = att_writer.key($ak)?;
-                                    let fun: &dyn Fn(&mut $crate::AttValueWriter) -> Result<(), std::fmt::Error> = &$av;
-                                    fun(&mut att_val_writer)?;
-                                    att_writer = att_val_writer.end()?;
-                                };
-                                ($ak:literal=$av:expr) => {
-                                    att_writer.write_attr($ak, $av)?;
-                                };
-                            }
-                            $d(attr!($k$t$v);)*
-
-                            macro_rules! content {
-                                ($content:block) => {
-                                    $writer = att_writer.close()?;
-                                    $content
-                                    $writer.write_end_tag(constructor.tag_name())?;
-                                };
-                                () => {
-                                    #[allow(unused_assignments)]
-                                    {
-                                        $writer = att_writer.close_empty()?;
-                                    }
-                                }
-                            }
-                            content!($d($blk)?);
-                        };
-                    }
-                }
+            #[allow(unused_assignments)]
+            {
+                $writer = att_writer.close_empty()?;
             }
-            $block
-        }
+        };
+    };
+    ($writer:ident, $constructor:expr $(, $k:literal$t:tt$v:expr)*, $content:block) => {
+        {
+            let constructor: &dyn $crate::TagConstructor = &$constructor;
+            #[allow(unused_mut)]
+            let mut att_writer = constructor.attributes($writer.open_start_tag(constructor.tag_name())?)?;
+
+            #[allow(unused_macros)]
+            macro_rules! attr {
+                ($ak:literal << $av:expr) => {
+                    let mut att_val_writer = att_writer.key($ak)?;
+                    let fun: &dyn Fn(&mut $crate::AttValueWriter) -> Result<(), std::fmt::Error> = &$av;
+                    fun(&mut att_val_writer)?;
+                    att_writer = att_val_writer.end()?;
+                };
+                ($ak:literal=$av:expr) => {
+                    att_writer.write_attr($ak, $av)?;
+                };
+            }
+            $(attr!($k$t$v);)*
+
+            $writer = att_writer.close()?;
+            $content
+            $writer.write_end_tag(constructor.tag_name())?;
+        };
     };
 }
 
@@ -220,11 +212,9 @@ mod tests {
 
     #[test]
     fn test_empty_tag() {
-        fn generate(w: XmlWriter) -> Result<(), std::fmt::Error> {
-            with_writer!(w, {
-                tag!("example");
-                tag!("example", "x" = 33);
-            });
+        fn generate(mut w: XmlWriter) -> Result<(), std::fmt::Error> {
+            tag!(w, "example");
+            tag!(w, "example", "x" = 33);
             Ok(())
         }
         let mut out = String::new();
@@ -238,10 +228,8 @@ mod tests {
             value.write("from function")?;
             Ok(())
         }
-        fn generate(w: XmlWriter) -> Result<(), Error> {
-            with_writer!(w, {
-                tag!("example", "key" << gen_value);
-            });
+        fn generate(mut w: XmlWriter) -> Result<(), Error> {
+            tag!(w, "example", "key" << gen_value);
             Ok(())
         }
         let mut out = String::new();
